"""Top-level package."""

__author__ = """Thomas Bouche, Johann Martin, Nicolas Roux"""
__email__ = "thomas.bouche@maif.fr"

from eurybia.core.smartdrift import SmartDrift

from .__version__ import __version__

__all__ = ["SmartDrift"]
